import csv
import logging
import os
from multiprocessing import Process
from multiprocessing.queues import Queue
from queue import Empty
from typing import List, Iterable

import redis

from galileodb.db import ExperimentDatabase
from galileodb.model import ServiceRequestTrace
from galileodb.sql.adapter import ExperimentSQLDatabase

logger = logging.getLogger(__name__)

POISON = "__POISON__"
START = "__START__"
PAUSE = "__PAUSE__"
FLUSH = '__FLUSH__'


def mkdirp(path):
    if not os.path.exists(path):
        os.makedirs(path)

    if os.path.isfile(path):
        raise FileExistsError("%s is an existing file" % path)


class TraceLogger(Process):
    flush_interval = 20

    def __init__(self, trace_queue: Queue, start=True) -> None:
        super().__init__()
        self.traces = trace_queue
        self.closed = False
        self.buffer = list()
        self.running = start

    def run(self):
        try:
            return self._loop()
        finally:
            self.flush()

    def flush(self):
        if not self.buffer:
            logger.debug('Buffer empty, not flushing')
            return

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Flushing trace buffer')

        self._do_flush(self.buffer)

        self.buffer.clear()

    def close(self):
        self.closed = True
        self.traces.put(POISON)

    def _loop(self):
        timeout = None
        while True:
            if self.closed and timeout is None:
                logger.debug('setting read timeout to 2 seconds')
                timeout = 2

            try:
                trace = self.traces.get(timeout=timeout)

                if trace == POISON:
                    logger.debug('poison received, setting closed to true')
                    self.closed = True
                    break
                elif trace == FLUSH:
                    logger.debug('flush command received, flushing buffer')
                    self.flush()
                    continue
                elif trace == START:
                    logger.debug('start received')
                    self.running = True
                    continue
                elif trace == PAUSE:
                    logger.debug('pause received, flushing remaining traces')
                    self.running = False
                    self.flush()
                    continue

                if self.running:
                    self.buffer.append(trace)

                if len(self.buffer) >= self.flush_interval:
                    logger.debug('flush interval reached, flushing buffer')
                    self.flush()

            except KeyboardInterrupt:
                break
            except Empty:
                logger.debug('queue is empty, exitting')
                return

    def _do_flush(self, buffer: List[ServiceRequestTrace]):
        pass


class TraceRedisLogger(TraceLogger):
    key = 'galileo:results:traces'

    def __init__(self, trace_queue: Queue, rds: redis.Redis, start=True) -> None:
        super().__init__(trace_queue, start)
        self.rds = rds

    def _do_flush(self, buffer: Iterable[ServiceRequestTrace]):
        rds = self.rds.pipeline()

        for trace in buffer:
            score = trace.created
            # FIXME
            if len(trace) == 6:
                value = '%s,%s,%s,%.7f,%.7f,%.7f' % trace
            elif len(trace) == 7:
                value = '%s,%s,%s,%.7f,%.7f,%.7f,%s' % trace

            rds.zadd(self.key, {value: score})

        rds.execute()


class TraceDatabaseLogger(TraceLogger):

    def __init__(self, trace_queue: Queue, experiment_db: ExperimentDatabase, start=True) -> None:
        super().__init__(trace_queue, start)
        self.experiment_db = experiment_db

    def run(self):
        if isinstance(self.experiment_db, ExperimentSQLDatabase):
            # this is a terrible hack due to multiprocessing issues:
            # close() will delete the threadlocal (which is not actually accessible from the process) and create a new
            # connection. The SqlAdapter adapter design may be broken. or python multiprocessing...
            self.experiment_db.db.reconnect()
        super().run()

    def _do_flush(self, buffer: Iterable[ServiceRequestTrace]):
        self.experiment_db.save_traces(list(buffer))


class TraceFileLogger(TraceLogger):

    def __init__(self, trace_queue: Queue, host_name, target_dir='/tmp/mc2/exp', start=True) -> None:
        super().__init__(trace_queue, start)
        self.target_dir = target_dir
        self.file_name = 'traces-%s.csv' % host_name
        self.file_path = os.path.join(self.target_dir, self.file_name)
        mkdirp(self.target_dir)

        self.init_file()

    def init_file(self):
        logger.debug('Initializing trace file logger to log into %s', self.file_path)
        if os.path.exists(self.file_path):
            return

        logger.debug('Initializing %s with header', self.file_path)
        with open(self.file_path, 'w') as fd:
            csv.writer(fd).writerow(ServiceRequestTrace._fields)

    def _do_flush(self, buffer: Iterable[ServiceRequestTrace]):
        with open(self.file_path, 'a') as fd:
            writer = csv.writer(fd)
            for row in buffer:
                writer.writerow(row)



import logging
import os
from typing import MutableMapping

from galileodb.db import ExperimentDatabase

logger = logging.getLogger(__name__)


def create_experiment_database_from_env(env: MutableMapping = os.environ) -> ExperimentDatabase:
    driver = env.get('galileo_expdb_driver', 'sqlite')
    return create_experiment_database(driver)


def create_experiment_database(driver: str, env: MutableMapping = os.environ) -> ExperimentDatabase:
    from galileodb.sql.adapter import ExperimentSQLDatabase

    if driver == 'sqlite':
        db_adapter = create_sqlite_from_env(env)

    elif driver == 'mysql':
        db_adapter = create_mysql_from_env(env)

    else:
        raise ValueError('unknown database driver %s' % driver)

    return ExperimentSQLDatabase(db_adapter)


def create_mysql_from_env(env: MutableMapping = os.environ):
    from galileodb.sql.driver.mysql import MysqlAdapter
    params = {
        'host': env.get('galileo_expdb_mysql_host', 'localhost'),
        'port': int(env.get('galileo_expdb_mysql_port', '3307')),
        'user': env.get('galileo_expdb_mysql_user', None),
        'password': env.get('galileo_expdb_mysql_password', None),
        'db': env.get('galileo_expdb_mysql_db', None)
    }

    logger.info('read mysql adapter parameters from environment %s', params)
    return MysqlAdapter(**params)


def create_sqlite_from_env(env: MutableMapping = os.environ):
    from galileodb.sql.driver.sqlite import SqliteAdapter
    db_file = env.get('galileo_expdb_sqlite_path', './galileodb.sqlite')

    logger.info('creating db adapter to SQLite %s', os.path.realpath(db_file))
    return SqliteAdapter(db_file)



import abc

from galileodb import ExperimentDatabase, Experiment, Telemetry
from galileodb.model import ExperimentEvent, ServiceRequestTrace


class AbstractTestExperimentDatabase(abc.ABC):
    db: ExperimentDatabase

    def test_get_experiment_on_invalid_id(self):
        result = self.db.get_experiment('NOTEXISTS')
        self.assertIsNone(result)

    def test_save_and_get_experiment(self):
        expected = Experiment('expid1', 'test_experiment', 'unittest', 10, 100, 1, 'running')

        self.db.save_experiment(expected)

        actual = self.db.get_experiment('expid1')

        self.assertEqual('expid1', actual.id)
        self.assertEqual('test_experiment', actual.name)
        self.assertEqual('unittest', actual.creator)
        self.assertEqual(10., actual.start)
        self.assertEqual(100., actual.end)
        self.assertEqual(1., actual.created)
        self.assertEqual('running', actual.status)

    def test_update_experiment_and_get(self):
        exp = Experiment('expid8', 'test_experiment', 'unittest', 10, None, 1, 'running')
        self.db.save_experiment(exp)

        self.assertEqual(self.db.get_experiment('expid8').status, 'running')
        self.assertEqual(self.db.get_experiment('expid8').end, None)

        exp.status = 'finished'
        exp.end = 100

        self.db.update_experiment(exp)

        self.assertEqual(self.db.get_experiment('expid8').status, 'finished')
        self.assertEqual(self.db.get_experiment('expid8').end, 100)

    def test_save_and_get_telemetry(self):
        telemetry = [
            Telemetry(1, 'cpu', 'n1', 32, 'expid1'),
            Telemetry(2, 'cpu', 'n1', 33, 'expid1'),
            Telemetry(3, 'cpu', 'n1', 31, 'expid2'),
        ]

        self.db.save_telemetry(telemetry)

        actual = self.db.get_telemetry('expid1')
        self.assertEqual(2, len(actual))
        self.assertEqual(telemetry[0], actual[0])
        self.assertEqual(telemetry[1], actual[1])

        actual = self.db.get_telemetry('expid2')
        self.assertEqual(1, len(actual))
        self.assertEqual(telemetry[2], actual[0])

        actual = self.db.get_telemetry()
        self.assertEqual(3, len(actual))
        self.assertEqual(telemetry[0], actual[0])
        self.assertEqual(telemetry[1], actual[1])
        self.assertEqual(telemetry[2], actual[2])

    def test_save_and_touch_and_get_traces(self):
        traces = [
            ServiceRequestTrace('c1', 's1', 'h1', 1.1, 1.2, 1.3),
            ServiceRequestTrace('c2', 's2', 'h2', 2.1, 2.2, 2.3),
            ServiceRequestTrace('c3', 's1', 'h1', 3.1, 3.2, 3.3),
            ServiceRequestTrace('c1', 's1', 'h1', 4.1, 4.2, 4.3),
        ]

        self.db.save_experiment(Experiment('exp1', start=1, end=3.5, status='FINISHED'))

        self.db.save_traces(traces)

        self.db.touch_traces(self.db.get_experiment('exp1'))  # should touch the first 3

        actual = self.db.get_traces('exp1')

        expected = [
            ServiceRequestTrace('c1', 's1', 'h1', 1.1, 1.2, 1.3, 'exp1'),
            ServiceRequestTrace('c2', 's2', 'h2', 2.1, 2.2, 2.3, 'exp1'),
            ServiceRequestTrace('c3', 's1', 'h1', 3.1, 3.2, 3.3, 'exp1'),
        ]

        self.assertEqual(expected, actual)

    def test_save_and_get_events(self):
        events = [
            ExperimentEvent('exp1', 1, 'begin'),
            ExperimentEvent('exp1', 2, 'start', 'function1'),
            ExperimentEvent('exp2', 3, 'stop', 'function1'),
        ]

        self.db.save_events(events)

        stored = self.db.get_events()
        self.assertEqual(3, len(stored))

        self.assertEqual(ExperimentEvent('exp1', 1.0, 'begin', None), stored[0])
        self.assertEqual(ExperimentEvent('exp1', 2.0, 'start', 'function1'), stored[1])
        self.assertEqual(ExperimentEvent('exp2', 3.0, 'stop', 'function1'), stored[2])

    def test_save_and_get_events_for_experiment(self):
        events = [
            ExperimentEvent('exp1', 1, 'begin'),
            ExperimentEvent('exp1', 2, 'start', 'function1'),
            ExperimentEvent('exp2', 3, 'stop', 'function1'),
        ]

        self.db.save_events(events)

        stored = self.db.get_events('exp1')
        self.assertEqual(2, len(stored))

        self.assertEqual(ExperimentEvent('exp1', 1.0, 'begin', None), stored[0])
        self.assertEqual(ExperimentEvent('exp1', 2.0, 'start', 'function1'), stored[1])

    def test_delete_experiment(self):
        exp_id = 'expid10'
        exp_id_control = 'expid11'
        self.db.save_experiment(Experiment(exp_id, 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_telemetry([Telemetry(1, 'cpu', 'n1', 32, exp_id)])

        self.db.save_experiment(Experiment(exp_id_control, 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_telemetry([Telemetry(1, 'cpu', 'n1', 32, exp_id_control)])

        self.assertIsNotNone(self.db.get_experiment(exp_id))

        self.db.delete_experiment(exp_id)

        self.assertIsNone(self.db.get_experiment(exp_id))
        self.assertIsNotNone(self.db.get_experiment(exp_id_control))
